<?php

declare(strict_types=1);

namespace Drupal\lms\Plugin\Field\FieldType;

use Drupal\Core\Cache\CacheableDependencyInterface;
use Drupal\Core\Cache\CacheableDependencyTrait;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Field\FieldItemList;
use Drupal\Core\Link;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\TypedData\ComputedItemListTrait;
use Drupal\Core\Url;
use Drupal\lms\Plugin\Block\CourseStartLink;

/**
 * Field item list class for group's 'start_link' field.
 *
 * @see lms_entity_bundle_field_info()
 */
final class StartLinkFieldItemList extends FieldItemList implements CacheableDependencyInterface {

  use CacheableDependencyTrait;
  use ComputedItemListTrait;

  /**
   * {@inheritdoc}
   *
   * Phpstan doesn't work well with abstract method in a trait.
   *
   * @phpstan-ignore-next-line
   */
  protected function computeValue(): void {
    $build = CourseStartLink::getRenderable();
    if (\count($build) === 0) {
      return;
    }

    $this->setCacheability(CacheableMetadata::createFromRenderArray($build));
    $renderer = \Drupal::service('renderer');
    $this->list[0] = $this->createItem(0, $renderer->render($build));
  }

  /**
   * Builds the themed action info.
   *
   * @param string $class
   *   The CSS class.
   * @param \Drupal\Core\StringTranslation\TranslatableMarkup $info
   *   Message to be displayed to end-user.
   *
   * @return array
   *   Render array.
   *
   * @todo Remove in the next major.
   */
  public static function buildActionInfo(string $class, TranslatableMarkup $info): array {
    return [
      '#type' => 'component',
      '#component' => 'lms:course_action_info',
      '#props' => [
        'class' => $class,
        'info_text' => $info,
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function view($display_options = []) {
    // Somehow on layout edition when an entity is not available,
    // the generateSampleItems() method is not called so we have
    // to call it here for new entities to have a proper preview.
    if ($this->getEntity()->isNew()) {
      $this->generateSampleItems(1);
    }
    return ['#markup' => $this->first()->getString()];
  }

  /**
   * {@inheritdoc}
   */
  public function generateSampleItems($count = 1): void {
    $renderer = \Drupal::service('renderer');
    for ($delta = 0; $delta < $count; $delta++) {
      $url = Url::fromRoute('<front>');

      $build = [
        '#type' => 'component',
        '#component' => 'lms:start_link',
        '#props' => [
          'link' => Link::fromTextAndUrl($this->t('Example link'), $url)->toRenderable(),
          'action_info' => NULL,
        ],
      ];
      $this->list[$delta] = $this->createItem(0, $renderer->render($build));
    }
  }

}
